/*
 * Using the Twitter STREAM API, filtering incoming tweets, 
 * saving them as .json objects or sending them as OSC messages.
 *
 * Three sets/filters are used in this version.
 *
 * @author Jon Bellona
 * @link http://jpbellona.com
 * @version 1.2.2 stand alone version, for public
 */

//in order to get OSC working, you'll need to cut and paste Kyma's IP here. 
var kymaip = "169.254.44.70"; //kyma

var util = require("util");
var fs = require("fs");
var connect = require('connect');
var serveStatic = require('serve-static');
var port = 9010;
var io = require('socket.io').listen(9011);

//osc shit
var osc = require('osc-min');
var dgram = require('dgram');
var udp = dgram.createSocket('udp4');
var outportk = 8000; //kyma
var outport = 41234; //max (doubled osc output for fun)

//midi
//https://github.com/justinlatimer/node-midi
//works great, but no clean access to midi channel!
//using easymidi instead.
var midi = require('easymidi'); //v0.0.2  https://www.npmjs.com/package/easymidi
var midiOut = new midi.Output('Cfeed', true); //virtual port that will be seen by Kyma Connect


// twitter
var twitter = require('ntwitter');

//@user information, API key goes here. The code won't work until you enter in your API data.
//go to https://apps.twitter.com/app/new to create your own API key.
var twit = new twitter({
  consumer_key: 'Enter_Your_Consumer_Key_Here',
  consumer_secret: 'Enter_Your_Consumer_Secret_Here',
  access_token_key: 'Enter_Your_Acess_Token_Key_Here',
  access_token_secret: 'Enter_Your_Acess_Token_Secret_Here'
});
var twitname = 'yourTwitterName'; //eg. ColbertAtHome


//the set to track. ***** Change your set number here!!! *******
var setNumber = 1; //i keep deleting sets, so pretty much setNumber = 1
var multiLocations = 1; //no location use 0; single location use 1, multi-locations 2
console.log("Running set number " + setNumber);
var canSendGeo = false;
var canSend = false;
var instrument = 0; //the instrument number is to group hashtags for a timbre (osc or midi channel)
var numInstruments = 5;

//instrument number indicates midi channel
//can include text not just hashtags.
//group is always on.
var set = {
  1: {
    name: 'Kyma Tweets',
    filter: {
      'track': 'carbonfeed, #c02feed, #co2feed, kyma, kiss2015, #carbon, #climate, #climatechange, #energy, natural gas, electricity, #environment, #earth, #green, #fracking, fruits, #water, trees',
      'locations': '-72.78, 43.202, -71.78, 44.202'
    },
    match_hashtags: {
      1: {
        instrument: 1,
        hashtag: ['carbonfeed', 'c02feed', 'co2feed', 'kyma', 'kiss2015']
      },
      2: {
        instrument: 2,
        hashtag: ['climate', 'climatechange', 'carbon']
      },
      3: {
        instrument: 4,
        hashtag: ['energy', 'fracking', 'electricity']
      }
    },
    match_text: {
      1: {
        instrument: 1,
        text: ['carbonfeed', 'c02feed', 'co2feed', 'kyma', 'kiss2015']
      },
      2: {
        instrument: 3,
        text: ['natural gas', 'electricity']
      },
      3: {
        instrument: 5,
        text: ['fruits', 'trees']
      }
    },
    match_locations: {
      1: {
        instrument: 2,
        coordinates: [-111.542934, 45.176998, -110.542934, 46.176998],
        name: 'Bozeman, MT region'
      }
    }
  }
};

//get lengths dynamically so we don't have to change our code below
var tweet_hashtags_length = Object.keys(set[setNumber]['match_hashtags']).length;
var tweet_locations_length = Object.keys(set[setNumber]['match_locations']).length;
var tweet_text_length = Object.keys(set[setNumber]['match_text']).length;


////// Locations have SW corner first (Long, Lat) //////
//LA: -118.491669, 33.584879, -117.807770, 34.168636
//NYC: -74,40,-73,41
//SF: -122.75,36.8,-121.75,37.8
//Moscow: 37, 55, 38, 56
//Iceland: -23.840332, 63.292939, -13.029785, 66.826520
//Cville region: -79,37.5,-78,38.5 
//Tokyo: 139.2, 35, 140.2, 36
//Bozeman, MT: -111.542934, 45.176998, -110.542934, 46.176998

//connect @v3.1.1 & serve-static @v1.5.3
//http://stackoverflow.com/questions/24346161/nodejs-connect-cannot-find-static
var app = connect();
app.use(serveStatic('public'));
app.listen(port);
util.log('the server is running on port: ' + port);

///////////////////////
/////////////////////// TWITTER STREAM
///////////////////////

//ntwitter setup+function
twit.stream('statuses/filter', set[setNumber].filter, function(stream) {
  stream.on('data', function (data) {

  //some tweets might be geotagged AND have the #hashtag that we're after.
  //filter to minimize any doubling.  (geotag supersedes hashtag.. e.g. overwrites instrument #)
  
  //setup data for OSC block
  var tweetText = data["text"];
  //console.log(tweetText); //write tweet text to Terminal.
  //console.log(""); //create space between tweets
  var tweetLen = tweetText.length; //0-142
  var tweetSum = 0;
  for (i = 0; i < tweetLen; i++) {
    tweetSum += tweetText.charCodeAt(i);
  }
  var tweetMean = tweetSum/tweetLen;
  var utc = 0;
  var utcPan = 0;
  if ((typeof(data["user"]["utc_offset"]) !== 'undefined') && (data["user"]["utc_offset"] != null)) {
    utc = data["user"]["utc_offset"] / 3600; //utc comes in seconds, not hours.
    utcPan = (utc + 12) / 26; //generate a value between 0-1
  } else {
    utc = getRandomInt(-10,12); //simulate utc timecodes
    utcPan = (utc + 12) / 26; //generate a value between 0-1
  }
  var statLen = data["user"]["statuses_count"];
  var statStrength = (statLen / 5000);
  statStrength = (statStrength > 1)? 1 : statStrength; 
  var tweetLang = data["lang"]; //language
  var tweetLangSum = tweetLang.charCodeAt(0) + tweetLang.charCodeAt(1);
  var langIndex0 = tweetLang.charCodeAt(0) % 5; //returns 0-4
  var langIndex1 = tweetLang.charCodeAt(1) % 5; //returns 0-4
  
  canSend = true; //only send once.
  instrument = 0; //reset instrument to 0, ignores errorneous tweets that don't fit criteria.


  ///////////////////////  OSC DATA ///////////////////////
  ///////////////////////           ///////////////////////
  
  //setup our OSC array to send out. 
  var tweetDataOSC = new Array();
  tweetDataOSC["/tweetText"] = tweetText;
  tweetDataOSC["/tweetLength"] = tweetLen;
  tweetDataOSC["/tweetSum"] = tweetSum;
  tweetDataOSC["/tweetMean"] = tweetMean;
  tweetDataOSC["/utc"] = utc;
  tweetDataOSC["/utcPan"] = utcPan;
  tweetDataOSC["/statuses"] = statLen;
  tweetDataOSC["/statusStrength"] = statStrength; //0-1
  tweetDataOSC["/tweetLang"] = tweetLang;
  tweetDataOSC["/tweetLangSum"] = tweetLangSum;
  tweetDataOSC["/instrument"] = instrument; //set this later.

  //////////////////////  MIDI DATA ///////////////////////
  ///////////////////////           ///////////////////////
  
  utcPitchArray = {
    '-12': 48,
    '-11': 50,
    '-10': 40,
    '-9': 41,
    '-8': 43,
    '-7': 48,
    '-6': 50,
    '-5': 52,
    '-4': 53,
    '-3': 55,
    '-2': 60,
    '-1': 62,
    '0': 64,
    '1': 65,
    '2': 67,
    '3': 72,
    '4': 74,
    '5': 76,
    '6': 77,
    '7': 79,
    '8': 84,
    '9': 86,
    '10': 88,
    '11': 89,
    '12': 91,
    '13': 60,
    '14': 48
  };

  //set up pitch classes here.
  var diatonicPitches = new Array(0, 2, 4, 5, 7, 9, 11);
  var pentaPitches = new Array(0, 2, 5, 7, 9);
  var languageOffsets1 = new Array(0, 3, 0, 7, 0); //0 0 3 2 7
  var languageOffsets2 = new Array(0, 0, 3, 2, 7); 
  var lengthPitchChromatic = scaleNum(tweetLen, 10, 142, 48, 72);
  
  /////////////////////// HASHTAGS ///////////////////////
  ///////////////////////          ///////////////////////

  //filter hashtags to trigger appropriate air instrument. (some are based upon geolocation)
  if ((typeof(data["entities"]["hashtags"]) !== 'undefined') && (data["entities"]["hashtags"] != null)) {

    for( var hashIndex = 1; hashIndex <= tweet_hashtags_length; hashIndex++ ) {
      for ( var hashIndex2 = Object.keys(set[setNumber]['match_hashtags'][hashIndex]['hashtag']).length -1; hashIndex2 >= 0; hashIndex2--) {
        var patt = new RegExp(set[setNumber]["match_hashtags"][hashIndex]['hashtag'][hashIndex2], "gi"); //create pattern based upon our set object

        //for each hashtag on this tweet, test our pattern.
        for (var i = data["entities"]["hashtags"].length - 1; i >= 0; i--) {
          if ( patt.test(data["entities"]["hashtags"][i]["text"])) {

            instrument = set[setNumber]["match_hashtags"][hashIndex]['instrument']; //match. set our group
            
            console.log(instrument);
            console.log(tweetText);
            console.log("");
            //send OSC & MIDI
            if ((instrument > 0) && (canSend)) { // && (canSend) && (instrument <= numInstruments)
              tweetDataOSC["/" + instrument + "/tweetText"] = tweetText;
              tweetDataOSC["/" + instrument + "/tweetLength"] = tweetLen;
              tweetDataOSC["/" + instrument + "/tweetSum"] = tweetSum;
              tweetDataOSC["/" + instrument + "/tweetMean"] = tweetMean;
              tweetDataOSC["/" + instrument + "/utc"] = utc;
              tweetDataOSC["/" + instrument + "/utcPan"] = utcPan;
              tweetDataOSC["/" + instrument + "/statuses"] = statLen;
              tweetDataOSC["/" + instrument + "/statusStrength"] = statStrength; //0-1
              tweetDataOSC["/" + instrument + "/tweetLang"] = tweetLang;
              tweetDataOSC["/" + instrument + "/tweetLangSum"] = tweetLangSum;
              tweetDataOSC["/instrument"] = instrument;

              sendOSCBundle(tweetDataOSC); //send OSC data
              sendShortOSC(tweetDataOSC["/instrument"], tweetDataOSC["/tweetText"]); //processing
              
              //send MIDI
              if (instrument <= numInstruments) {
                sendMIDInote(utcPitchArray[utc.toString()], 120, scaleNum(tweetLen, 0, 140, 400, 600), instrument);
              }
              //send another MIDI note
              if (instrument != 3) {
                sendMIDInote(lengthPitchChromatic, scaleNum(statStrength, 0, 1, 80, 127), scaleNum(tweetLen, 0, 140, 400, 600), instrument);
              }
              canSend = false;
            }
          }
        }
        
      }
    } 
  }

  /////////////////////// GEO-TAGGED ///////////////////////
  ///////////////////////            ///////////////////////

  //check for geotagged tweets. (e.g. #gaga could be geotagged but outside LA)
  var geolocation = new Array();
  if ((typeof(data["geo"]) !== 'undefined') && (data["geo"] != null)) {
    //console.log(data["geo"]);
    geolocation[0] = data["geo"]["coordinates"][0]; //lat
    geolocation[1] = data["geo"]["coordinates"][1]; //long
    geolocation[2] = 'Anytown, USA';

    //based upon the hashtag set, filter the geotagged tweets
    //eventually put this into a single function.
    var swLat = 0;
    var neLat = 0;
    var swLong = 0;
    var neLong = 0;
    switch(multiLocations) {
      case 1:
        //location data (based upon a single locale [1])
        swLat = set[setNumber]['match_locations'][1]['coordinates'][1];
        neLat = set[setNumber]['match_locations'][1]['coordinates'][3];
        swLong = set[setNumber]['match_locations'][1]['coordinates'][0];
        neLong = set[setNumber]['match_locations'][1]['coordinates'][2];
        if ( (geolocation[0] > swLat) && (geolocation[0] < neLat) ) {
          if ( (geolocation[1] > swLong) && (geolocation[1] < neLong) ) {
            geolocation[2] = set[setNumber]['match_locations'][1]['name'];
            instrument = set[setNumber]['match_locations'][1]['instrument']; //5
            canSendGeo = true;
            break;
          }
        }
        break;
      case 2:
        //location data (based upon a multiple locations [x])
        for( var loc = 1; loc <= tweet_locations_length; loc++ ) {
          swLat = set[setNumber]['match_locations'][loc]['coordinates'][1];
          neLat = set[setNumber]['match_locations'][loc]['coordinates'][3];
          swLong = set[setNumber]['match_locations'][loc]['coordinates'][0];
          neLong = set[setNumber]['match_locations'][loc]['coordinates'][2];
          if ( (geolocation[0] > swLat) && (geolocation[0] < neLat) ) {
            if ( (geolocation[1] > swLong) && (geolocation[1] < neLong) ) {
              geolocation[2] = set[setNumber]['match_locations'][loc]['name'];
              instrument = set[setNumber]['match_locations'][loc]['instrument']; //an integer, e.g. 3
              canSendGeo = true;
              break;
            }
          }
        }
      default:
        canSendGeo = false;
    }

    //set the instrument's pitch
    if ((instrument > 0) && (instrument <= numInstruments) && (canSendGeo = true)) {
      tweetDataOSC["/instrument"] = instrument;
    }

  } else {
    //send non geotagged tweet (b/c sometimes comes up with instrument 0)
    if (instrument > 0 ) {
      tweetDataOSC["/instrument"] = instrument;
    }
  }

  if ((instrument > 0) && (instrument <= numInstruments) && (canSend != false)) {

    console.log(instrument + " geo");
    console.log(tweetText);
    console.log("");

    tweetDataOSC["/" + instrument + "/tweetText"] = tweetText;
    tweetDataOSC["/" + instrument + "/tweetLength"] = tweetLen;
    tweetDataOSC["/" + instrument + "/tweetSum"] = tweetSum;
    tweetDataOSC["/" + instrument + "/tweetMean"] = tweetMean;
    tweetDataOSC["/" + instrument + "/utc"] = utc;
    tweetDataOSC["/" + instrument + "/utcPan"] = utcPan;
    tweetDataOSC["/" + instrument + "/statuses"] = statLen;
    tweetDataOSC["/" + instrument + "/statusStrength"] = statStrength; //0-1
    tweetDataOSC["/" + instrument + "/tweetLang"] = tweetLang;
    tweetDataOSC["/" + instrument + "/tweetLangSum"] = tweetLangSum;
    tweetDataOSC["/instrument"] = instrument;

    sendOSCBundle(tweetDataOSC); //send OSC data
    sendShortOSC(tweetDataOSC["/instrument"], tweetDataOSC["/tweetText"]);
    canSend = false;

    //send MIDI
    if (instrument <= numInstruments) {
      sendMIDInote(pentaPitches[tweetLen%5]+24, 120, scaleNum(tweetLen, 0, 140, 600, 900), instrument);
    }
    //send another MIDI note, chromatic
    if (instrument != 3) {
      sendMIDInote(lengthPitchChromatic, scaleNum(statStrength, 0, 1, 80, 127), scaleNum(tweetLen, 0, 140, 400, 600), instrument);
    }
  }

  ///////////////////// TWEET TEXT ///////////////////////
  /////////////////////            ///////////////////////

  for( var txtIndex = 1; txtIndex <= tweet_text_length; txtIndex++ ) {
    for ( var txt2 = Object.keys(set[setNumber]['match_text'][txtIndex]['text']).length -1; txt2 >= 0; txt2--) {
      var patt = new RegExp(set[setNumber]["match_text"][txtIndex]['text'][txt2], "gi"); //create pattern based upon our set object
      if ( patt.test(tweetText)) {
        instrument = set[setNumber]["match_text"][txtIndex]['instrument']; //match sets our group

        console.log(instrument + " text");
        console.log(tweetText);
        console.log("");
        //then send out.
        if ((instrument > 0) && (instrument <= numInstruments) && (canSend != false)) {
          tweetDataOSC["/" + instrument + "/tweetText"] = tweetText;
          tweetDataOSC["/" + instrument + "/tweetLength"] = tweetLen;
          tweetDataOSC["/" + instrument + "/tweetSum"] = tweetSum;
          tweetDataOSC["/" + instrument + "/tweetMean"] = tweetMean;
          tweetDataOSC["/" + instrument + "/utc"] = utc;
          tweetDataOSC["/" + instrument + "/utcPan"] = utcPan;
          tweetDataOSC["/" + instrument + "/statuses"] = statLen;
          tweetDataOSC["/" + instrument + "/statusStrength"] = statStrength; //0-1
          tweetDataOSC["/" + instrument + "/tweetLang"] = tweetLang;
          tweetDataOSC["/" + instrument + "/tweetLangSum"] = tweetLangSum;
          tweetDataOSC["/instrument"] = instrument;

          sendOSCBundle(tweetDataOSC); //send OSC data
          sendShortOSC(tweetDataOSC["/instrument"], tweetDataOSC["/tweetText"]);
          canSend = false;
          //send MIDI
          if (instrument <= numInstruments) {
            sendMIDInote(utcPitchArray[utc.toString()], 120, scaleNum(tweetLen, 0, 140, 400, 600), instrument);
          } else {
            sendMIDInote(pentaPitches[tweetLen%5]+24, 120, scaleNum(tweetLen, 0, 140, 600, 900), instrument);
          }
        }
      }
    }
  } 
	
  });
});


/**
 * Callback from sendMIDInote, in order to turn note off (so no hanging notes!)
 * It is possible to get hanging notes if turning off node with Kyma running in bg.
 */
function noteOff(note, chan){
  midiOut.send( 'noteoff', {
    note: note,
    velocity: 0,
    channel: chan
  });
}

/**
 * Send out MIDI note based upon tweet features.
 * @param  {[type]} note pitch based upon tweet length
 * @param  {[type]} velo velocity is static currently
 * @param  {[type]} dur  duration is static currently
 * @param  {[type]} chan channel is human filter (the hashtag)
 */
function sendMIDInote(note, velo, dur, chan) {
  midiOut.send( 'noteon', {
    note: note,
    velocity: velo,
    channel: chan-1
  });
  setTimeout(noteOff, dur, note, chan-1); //turn note off after X ms
}

/**
 * Fire off test OSC and MIDI notes so don't have to be dependent upon Twitter
 */
function testingFunction(inst) {
  var tweetDataOSC = new Array();
  tweetDataOSC["/" + inst + "/tweetText"] = 'tweetText';
  tweetDataOSC["/" + inst + "/tweetLength"] = getRandomInt(10, 140);
  tweetDataOSC["/" + inst + "/tweetSum"] = getRandomInt(1000, 5000);
  tweetDataOSC["/" + inst + "/tweetMean"] = getRandomInt(75, 110);
  tweetDataOSC["/" + inst + "/utc"] = getRandomInt(-10, 12);
  tweetDataOSC["/" + inst + "/utcPan"] = (getRandomInt(-10, 12) + 12) / 26;
  tweetDataOSC["/" + inst + "/statuses"] = getRandomInt(100, 5000);
  var statStrTmp = getRandomInt(100, 5000) / 5000;
  tweetDataOSC["/" + inst + "/statusStrength"] = statStrTmp; //0-1
  tweetDataOSC["/" + inst + "/tweetLang"] = 'en';
  tweetDataOSC["/" + inst + "/tweetLangSum"] = getRandomInt(211, 230);
  tweetDataOSC["/instrument"] = inst;
  sendOSCBundle(tweetDataOSC); //send OSC data
  var nn = getRandomInt(48, 72);
  sendMIDInote(nn, 120, 200, inst);
  var nn2 = getRandomInt(48, 72);
  sendMIDInote(nn2, scaleNum(statStrTmp, 0, 1, 80, 127), 200, instrument);
}
//test intervals
// setInterval(testingFunction, 603, 2); //millisecond interval, instrument number
// setInterval(testingFunction, 581, 3);
// setInterval(testingFunction, 547, 4);
// setInterval(testingFunction, 543, 5);
// setInterval(testingFunction, 501, 6);
// setInterval(testingFunction, 432, 7);
// setInterval(testingFunction, 406, 8);
// setInterval(testingFunction, 330, 9);

/**
 * Send multidimensional array as an OSC bundle (keys are addresses)
 */
function sendOSCBundle(bundle) {
  var oscBundle = new Array();

  for (var k in bundle){
    if (bundle.hasOwnProperty(k)) {
      var tmpItem = { address: k, args: bundle[k] };
      oscBundle.push(tmpItem);
    }
  }

  var buf;
  buf = osc.toBuffer(
  {
    oscType: "bundle",
    elements:
     oscBundle
  });
  //udp.send(buf, 0, buf.length, outport, "localhost"); //max
    //typetags are sffffsff
  return udp.send(buf, 0, buf.length, outportk, kymaip); //kyma
};

/**
 * Send shortened OSC data out as OSC message to additional OSC output.
 *
 * @param {Array} [geolocation] lat, long, cityname
 * geolocation[0] = float latitude
 * geolocation[1] = float long
 * geolocation[2] = String name of city
 * @param {String} [tweetTxt] number of the tweetTxt to trigger
 */
sendShortOSC = function(id, tweetTxt) {
  var buf;
  buf = osc.toBuffer({
    address: "/tweet",
    args: [
      id, tweetTxt
    ]
  });
  return udp.send(buf, 0, buf.length, outport, "localhost");
};


// Returns a random integer between min (included) and max (excluded)
// Using Math.round() will give you a non-uniform distribution!
function getRandomInt(min, max) {
  return Math.floor(Math.random() * (max - min)) + min;
}

/**
 * Input a number and define input/output range.
 * similar to 'scale' object in Max/MSP.
 */
function scaleNum(input, iMin, iMax, oMin, oMax) {
  var percent = (input - iMin) / (iMax - iMin);
  var output = percent * (oMax - oMin) + oMin;
  return output;
}